/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
     
     Vulkan Graphics API: 11 Steps

     Minimalistic examples to build upon and experiment with 
     to develop an understanding of the underlying API

     
/*
/****************************************************************/
/*
   About:
   Self contained demo (i.e., each section encapsulated within
   a function - see the parameter dependencies). 
   Remember to include and compile the two binary shader
   files (vertex and shader)

   ''''
   #define for enabling/disabling `DEPTH_BUFFER' - 
   extension (additional code) is needed from the minimal
   version in the original book to have a working
   depth buffer (i.e., zbuffer).
   ''''

*/


#include "setup.h"

#include "texture.h"



#pragma comment(lib, "C:\\VulkanSDK\\1.0.17.0\\Bin32\\vulkan-1.lib")
// or static lib doesn't depend on dll
//#pragma comment(lib, "VKstatic.1.lib")


// Do we want to enable the added Vuklan debug?
#define ENABLE_VULKAN_DEBUG_CALLBACK





    
#ifdef ENABLE_VULKAN_DEBUG_CALLBACK // Debug callback
// Set this function as a debug callback when we initialize
// Vulkan to let us know if something went wrong
VKAPI_ATTR VkBool32 VKAPI_CALL 
  MyDebugReportCallback( VkDebugReportFlagsEXT flags, 
               VkDebugReportObjectTypeEXT objectType, 
               uint64_t object,
               size_t location, 
               int32_t messageCode, 
               const char* pLayerPrefix, 
               const char* pMessage, 
               void* pUserData ) 
{
    dprintf( pLayerPrefix );
    dprintf( " " );
    dprintf( pMessage );
    dprintf( "\n" );
    DBG_ASSERT(false);
    return VK_FALSE;
}// End MyDebugReportCallback(..)
#endif





// Win32 message callback
LRESULT CALLBACK WindowProc( HWND  hwnd, 
                             UINT  uMsg, 
                             WPARAM wParam, 
                             LPARAM lParam ) 
{
    switch( uMsg ) 
    {
        // called when we close the application
        case WM_CLOSE: 
        {
            // posts a WM_QUIT message
            PostQuitMessage( 0 );
        }
        break;
        
        // windows informs us that we need to redraw
        case WM_PAINT: 
        {
        }
        break;
        
        default: 
        {
            break;
        }
        break;
    }// End switch(..)
    
    // a pass-through for now. We will return to this 
    // callback
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}// End WindowProc(..)




/****************************************************************/
/*  
     Step 1
/*
/****************************************************************/

// Step 1 - Initializing Window
void SetupWindow(int    width, 
                 int    height, 
                 HWND*  outWindowHandle)
{
  // defines our window properties
  HINSTANCE hInst           = GetModuleHandle(NULL);
  WNDCLASSEX windowClass    = {};
  windowClass.cbSize        = sizeof(WNDCLASSEX);
  windowClass.style         = CS_OWNDC | CS_VREDRAW | CS_HREDRAW;
  windowClass.lpfnWndProc   = WindowProc;
  windowClass.hInstance     = hInst;
  windowClass.lpszClassName = "VulkanWindowClass";
  RegisterClassEx( &windowClass );

  *outWindowHandle = 
  CreateWindowEx( NULL, 
                  "VulkanWindowClass",
                  "Hello Vulkan",
                  WS_OVERLAPPEDWINDOW | WS_VISIBLE,
                  100, 
                  100, 
                  width,
                  height,
                  NULL,
                  NULL,
                  hInst,
                  NULL );
  DBG_ASSERT(*outWindowHandle!=NULL);    
}// End SetupWindow(..)



/****************************************************************/
/*  
     Step 2
/*
/****************************************************************/

// Step 2 - Initialize Vulkan  
void SetupVulkanInstance(HWND           windowHandle,
                         VkInstance*    outInstance,
                         VkSurfaceKHR*  outSurface)  
{
  // Initialize VULKAN

  // Layer properties
  uint32_t count = 0;
  // vkEnumerateInstanceLayerProperties - Returns number of global 
  // layer properties. Properties is either NULL or a pointer to 
  // an array of VkLayerProperties structures.
  VkResult result = 
  vkEnumerateInstanceLayerProperties(&count, // uint32_t*
                                     NULL);  // VkLayerProperties*

  DBG_ASSERT( VK_SUCCESS == result );
  DBG_ASSERT( count > 0 );

  vector<VkLayerProperties> instanceLayers;
  instanceLayers.resize(count);
  // As the VkLayerProperties structure array is not NULL the function returns
  // the layer properties.  
  result = 
  vkEnumerateInstanceLayerProperties(&count, &instanceLayers[0]);


  // Extension properties
  // The VkExtensionProperties count relates to the number of 
  // extension properties available
  result = vkEnumerateInstanceExtensionProperties(NULL, 
                                                  &count, 
                                                  NULL);
  
  DBG_ASSERT(VK_SUCCESS==result);
  DBG_ASSERT(count > 0);

  vector<VkExtensionProperties> instanceExtension;
  instanceExtension.resize(count);
  // Array of LayerNames not  NULL so returns an array of null-terminated UTF-8 strings 
  // names for the retrievable extensions.
  // The VkExtensionProperties structures is not NULL so returns the extension properties
  result = vkEnumerateInstanceExtensionProperties(NULL, &count, &instanceExtension[0]);

  vector<string> extensionNames;
  extensionNames.resize(count);


  /*
    e.g., VK_LAYER_NV_optimus or VK_LAYER_LUNARG_standard_validation
  */

  const char *layers[] = { "VK_LAYER_NV_optimus" };

#ifdef ENABLE_VULKAN_DEBUG_CALLBACK // access debug callbacks
  const char *extensions[] = {  "VK_KHR_surface", 
                                "VK_KHR_win32_surface",
                                "VK_EXT_debug_report"};
  
#else
  const char *extensions[] = { "VK_KHR_surface", 
                               "VK_KHR_win32_surface" };
#endif

  {
  VkApplicationInfo ai    = { }; 
  ai.sType          = VK_STRUCTURE_TYPE_APPLICATION_INFO;
  ai.pApplicationName      = "Hello Vulkan";
  ai.engineVersion      = 1;
  ai.apiVersion        = VK_API_VERSION_1_0;

  VkInstanceCreateInfo ici  = { };
  ici.sType                 = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
  ici.flags                 = 0;
  ici.pNext                 = 0;
  ici.pApplicationInfo      = &ai;
  ici.enabledLayerCount     = 1;
  ici.ppEnabledLayerNames   = layers;
  ici.enabledExtensionCount = 2;
#ifdef ENABLE_VULKAN_DEBUG_CALLBACK // access debug callbacks
  ici.enabledExtensionCount = 3;
#endif
  ici.ppEnabledExtensionNames = extensions;

  VkResult result = 
  vkCreateInstance( &ici, 
                    NULL, 
                    outInstance );

  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create vulkan instance." );
  DBG_ASSERT(*outInstance!=NULL);
  }

  // Optional - if we want Vulkan to tell us if something is wrong
  // we must set the callback
#ifdef ENABLE_VULKAN_DEBUG_CALLBACK
  {
    // Register our error logging function (defined at the top of the file)
    VkDebugReportCallbackEXT error_callback   = VK_NULL_HANDLE;
    VkDebugReportCallbackEXT warning_callback = VK_NULL_HANDLE;

    PFN_vkCreateDebugReportCallbackEXT vkCreateDebugReportCallbackEXT = NULL;

    *(void **)&vkCreateDebugReportCallbackEXT = 
      vkGetInstanceProcAddr( *outInstance, "vkCreateDebugReportCallbackEXT" );
    DBG_ASSERT(vkCreateDebugReportCallbackEXT);


    VkDebugReportCallbackCreateInfoEXT cb_create_info = {};
    cb_create_info.sType       = VK_STRUCTURE_TYPE_DEBUG_REPORT_CREATE_INFO_EXT;
    cb_create_info.flags       = VK_DEBUG_REPORT_ERROR_BIT_EXT;
    cb_create_info.pfnCallback = &MyDebugReportCallback;

    VkResult result = 
    vkCreateDebugReportCallbackEXT(*outInstance, &cb_create_info,
                                     nullptr, &error_callback);
    DBG_ASSERT_VULKAN_MSG(result, "vkCreateDebugReportCallbackEXT(ERROR) failed");

    // Capture warning as well as errors
    cb_create_info.flags = VK_DEBUG_REPORT_WARNING_BIT_EXT |
                            VK_DEBUG_REPORT_PERFORMANCE_WARNING_BIT_EXT;
    cb_create_info.pfnCallback = &MyDebugReportCallback;

    result = 
    vkCreateDebugReportCallbackEXT(*outInstance, 
                                    &cb_create_info,
                                    nullptr, 
                                    &warning_callback);
    DBG_ASSERT_VULKAN_MSG(result, 
        "vkCreateDebugReportCallbackEXT(WARN) failed");
  }
#endif


  // We need to define what type of surface we'll be 
  // rendering to - this will depend on our computer 
  // and operating system

  HINSTANCE hInst  = GetModuleHandle(NULL);
  // setup parameters for our new windows 
  // surface we'll render into:
  VkWin32SurfaceCreateInfoKHR sci = {};
  sci.sType       = VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR;
  // parameter is NULL, GetModuleHandle returns a handle 
  // to the file used to create the calling process 
  sci.hinstance   = hInst;
  // Your `created' window handle (HWND)
  sci.hwnd        = windowHandle;

  DBG_ASSERT(*outSurface==NULL);
  // To create a VkSurfaceKHR object for a Win32 window, call:
  result = 
  vkCreateWin32SurfaceKHR( *outInstance, &sci, NULL, outSurface );

  DBG_ASSERT_VULKAN_MSG( result, 
    "Could not create surface." );
  DBG_ASSERT(outSurface!=NULL);
}// End SetupVulkanInstance(..)


/****************************************************************/
/*  
     Step 3
/*
/****************************************************************/

// Step 3 - Find/Create Device  
void SetupPhysicalDevice(VkInstance         instance,
                         VkPhysicalDevice*  outPhysicalDevice,
                         VkDevice*          outDevice)
{
  // Query how many devices are present in the system
  uint32_t deviceCount = 0;

  // Enumerates the physical devices accessible to a Vulkan instance
  // The instance is the handle to a Vulkan instance you previously 
  // created with vkCreateInstance.  The VkPhysicalDevice pointer 
  // can be either NULL or a pointer to an array of VkPhysicalDevice handles.
  VkResult result = 
  vkEnumeratePhysicalDevices(instance, 
                             &deviceCount, 
                             NULL);
  DBG_ASSERT_VULKAN_MSG(result, 
    "Failed to query the number of physical devices present");

  // There has to be at least one device present
  DBG_ASSERT_MSG(0 != deviceCount, 
    "Couldn't detect any device present with Vulkan support");

  // Get the physical devices
  vector<VkPhysicalDevice> physicalDevices(deviceCount);
  result = 
  vkEnumeratePhysicalDevices(instance, &deviceCount, &physicalDevices[0]);
  DBG_ASSERT_VULKAN_MSG(result, 
    "Faied to enumerate physical devices present");
  DBG_ASSERT(physicalDevices.size()>0);
  
  // Use the first available device
  *outPhysicalDevice = physicalDevices[0];


  // Enumerate all physical devices and print out the details
  for (uint32_t i = 0; i < deviceCount; ++i) 
  {
    VkPhysicalDeviceProperties deviceProperties;
    memset(&deviceProperties, 0, sizeof deviceProperties);
    vkGetPhysicalDeviceProperties(physicalDevices[i], &deviceProperties);
    dprintf("Driver Version: %d\n", deviceProperties.driverVersion);
    dprintf("Device Name:    %s\n", deviceProperties.deviceName);
    dprintf("Device Type:    %d\n", deviceProperties.deviceType);
    dprintf("API Version:    %d.%d.%d\n",
      (deviceProperties.apiVersion>>22)&0x3FF,
      (deviceProperties.apiVersion>>12)&0x3FF,
      (deviceProperties.apiVersion&0xFFF));
  }//End for i




  // Fill up the physical device memory properties: 
  VkPhysicalDeviceMemoryProperties memoryProperties;
  vkGetPhysicalDeviceMemoryProperties( *outPhysicalDevice, &memoryProperties );

  // Here's where we initialize our queues
  VkDeviceQueueCreateInfo queueCreateInfo = {};
  queueCreateInfo.sType       = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
  // Use the first queue family in the family list
  queueCreateInfo.queueFamilyIndex = 0;
  queueCreateInfo.queueCount       = 1;
  float queuePriorities[]          = { 1.0f };
  queueCreateInfo.pQueuePriorities = queuePriorities;

  const char *deviceExtensions[] = { "VK_KHR_swapchain" };

  const char *layers[] = { "VK_LAYER_NV_optimus" };

  VkDeviceCreateInfo dci = {};
  dci.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
  // Set queue(s) into the device
  dci.queueCreateInfoCount    = 1;
  dci.pQueueCreateInfos       = &queueCreateInfo;
  dci.enabledLayerCount       = 0;
  dci.ppEnabledLayerNames     = layers;
  dci.enabledExtensionCount   = 1;
  dci.ppEnabledExtensionNames = deviceExtensions;

  VkPhysicalDeviceFeatures features = {};
  features.shaderClipDistance = VK_TRUE;
  dci.pEnabledFeatures        = &features;

  // Ideally, we'd want to enumerate and find the best
  // device, however, we just use the first device 
  // `physicalDevices[0]' for our sample, which we
  // stored in the previous chapter
  result = 
    vkCreateDevice( *outPhysicalDevice, &dci, NULL, outDevice );
  DBG_ASSERT_VULKAN_MSG( result, "Failed to create logical device!" );
}// End SetupPhysicalDevice(..)










/****************************************************************/
/*  
     Step 4
/*
/****************************************************************/

// Step -4-
void SetupSwapChain(VkDevice          device,
                    VkPhysicalDevice  physicalDevice,
                    VkSurfaceKHR      surface,
                    int*              outWidth,
                    int*              outHeight,
                    VkSwapchainKHR*   outSwapChain,           
                    VkImage**         outPresentImages,             
                    VkImageView**     outPresentImageViews)
{
  {
  // Create swap-chain
  // swap-chain creation:
  VkSurfaceCapabilitiesKHR surfaceCapabilities = {};
  vkGetPhysicalDeviceSurfaceCapabilitiesKHR(  physicalDevice, 
                                              surface, 
                                              &surfaceCapabilities );

  VkExtent2D surfaceResolution = surfaceCapabilities.currentExtent;
  *outWidth             = surfaceResolution.width;
  *outHeight            = surfaceResolution.height;


  VkSwapchainCreateInfoKHR ssci = {};
  ssci.sType            = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR;
  ssci.surface          = surface;
  // We'll use 2 for `double' buffering
  ssci.minImageCount    = 2;
  ssci.imageFormat      = VK_FORMAT_B8G8R8A8_UNORM;
  ssci.imageColorSpace  = VK_COLORSPACE_SRGB_NONLINEAR_KHR;
  ssci.imageExtent      = surfaceResolution;
  ssci.imageArrayLayers = 1;
  ssci.imageUsage       = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
  ssci.imageSharingMode = VK_SHARING_MODE_EXCLUSIVE;
  ssci.preTransform     = VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR;
  ssci.compositeAlpha   = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
  ssci.presentMode      = VK_PRESENT_MODE_MAILBOX_KHR;
  // If we want clipping outside the extents
  ssci.clipped          = true;     
  ssci.oldSwapchain     = NULL;

  VkResult result = 
  vkCreateSwapchainKHR( device, 
                        &ssci, 
                        NULL, 
                        outSwapChain );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create swapchain." );
  }

  // Chapter 7
  // Create our images 'double' buffering
  {
  uint32_t imageCount    = 0;
  vkGetSwapchainImagesKHR( device, 
                           *outSwapChain, 
                           &imageCount, 
                           NULL );
  DBG_ASSERT(imageCount==2);

  // this should be 2 for double-buffering
  *outPresentImages = new VkImage[imageCount];  

  // link the images to the swapchain  
  VkResult result =
  vkGetSwapchainImagesKHR( device, 
                           *outSwapChain, 
                           &imageCount, 
                           *outPresentImages );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create swap-chain images" );
  }


  {
  // We have 2 for double buffering
  *outPresentImageViews = new VkImageView[2];
  for( uint32_t i = 0; i < 2; ++i ) 
  {
  // create VkImageViews for our swap chain 
  // VkImages buffers:
  VkImageViewCreateInfo ivci            = {};
  ivci.sType                            = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
  ivci.viewType                         = VK_IMAGE_VIEW_TYPE_2D;
  ivci.format                           = VK_FORMAT_B8G8R8A8_UNORM;
  ivci.components.r                     = VK_COMPONENT_SWIZZLE_R;
  ivci.components.g                     = VK_COMPONENT_SWIZZLE_G;
  ivci.components.b                     = VK_COMPONENT_SWIZZLE_B;
  ivci.components.a                     = VK_COMPONENT_SWIZZLE_A;
  ivci.subresourceRange.aspectMask      = VK_IMAGE_ASPECT_COLOR_BIT;
  ivci.subresourceRange.baseMipLevel    = 0;
  ivci.subresourceRange.levelCount      = 1;
  ivci.subresourceRange.baseArrayLayer  = 0;
  ivci.subresourceRange.layerCount      = 1;
  ivci.image                            = (*outPresentImages)[i];

  VkResult result = 
  vkCreateImageView(  device, 
            &ivci, 
            NULL, 
            &(*outPresentImageViews)[i] );

  DBG_ASSERT_VULKAN_MSG( result, 
    "Could not create ImageView." );
  }// End for i
  }
}// End SetupSwapChain(..)



/****************************************************************/
/*  
     Step 5
/*
/****************************************************************/

// Step -5-
void SetupRenderPass(VkDevice         device,
                    VkPhysicalDevice  physicalDevice,
                    int               width,
                    int               height,
                    VkImageView*      presentImageViews,
                    VkRenderPass*     outRenderPass, 
                    VkFramebuffer**   outFrameBuffers)
{
  // Frame buffer
  // define our attachment points 

#ifdef DEPTH_BUFFER
  // Extension (Depth Buffer)
  VkImage           depthImage       = NULL;
  VkImageView      depthImageView      = NULL;

  {
  // create a depth image:
  VkImageCreateInfo imageCreateInfo     = {};
  imageCreateInfo.sType                 = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
  imageCreateInfo.imageType             = VK_IMAGE_TYPE_2D;
  imageCreateInfo.format                = VK_FORMAT_D16_UNORM;
  VkExtent3D ef                         = { width, height, 1 };
  imageCreateInfo.extent                = ef;
  imageCreateInfo.mipLevels             = 1;
  imageCreateInfo.arrayLayers           = 1;
  imageCreateInfo.samples               = VK_SAMPLE_COUNT_1_BIT;
  imageCreateInfo.tiling                = VK_IMAGE_TILING_OPTIMAL;
  imageCreateInfo.usage                 = VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT;
  imageCreateInfo.sharingMode           = VK_SHARING_MODE_EXCLUSIVE;
  imageCreateInfo.queueFamilyIndexCount = 0;
  imageCreateInfo.pQueueFamilyIndices   = NULL;
  imageCreateInfo.initialLayout         = VK_IMAGE_LAYOUT_UNDEFINED;
  VkResult result = 
  vkCreateImage( device, &imageCreateInfo, NULL, &depthImage );
  DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to create depth image." );

  VkMemoryRequirements memoryRequirements = {};
  vkGetImageMemoryRequirements( device, depthImage, &memoryRequirements );

  // Allocate memory for our depth buffer
  VkMemoryAllocateInfo imageAllocateInfo  = {};
  imageAllocateInfo.sType                 = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
  imageAllocateInfo.allocationSize        = memoryRequirements.size;

  // memoryTypeBits is a bitfield where if bit i is set, it means that 
  // the VkMemoryType i of the VkPhysicalDeviceMemoryProperties structure 
  // satisfies the memory requirements:
  // read the device memory properties
  VkPhysicalDeviceMemoryProperties memoryProperties;
  vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

  uint32_t memoryTypeBits = memoryRequirements.memoryTypeBits;
  VkMemoryPropertyFlags desiredMemoryFlags = VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT;
  for( uint32_t i = 0; i < VK_MAX_MEMORY_TYPES; ++i ) 
  {
    VkMemoryType memoryType = memoryProperties.memoryTypes[i];
    if( memoryTypeBits & 1 ) 
    {
        if( ( memoryType.propertyFlags & desiredMemoryFlags ) == desiredMemoryFlags ) 
        {
                  imageAllocateInfo.memoryTypeIndex = i;
                  break;
        }
    }
    memoryTypeBits = memoryTypeBits >> 1;
  }

    VkDeviceMemory imageMemory = { 0 };
    result = vkAllocateMemory( device, &imageAllocateInfo, NULL, &imageMemory );
    DBG_ASSERT_VULKAN_MSG( result, "Failed to allocate device memory." );

    result = vkBindImageMemory( device, depthImage, imageMemory, 0 );
    DBG_ASSERT_VULKAN_MSG( result, "Failed to bind image memory." );

    // create the depth image view:
    VkImageAspectFlags aspectMask             = VK_IMAGE_ASPECT_DEPTH_BIT;
    VkImageViewCreateInfo imageViewCreateInfo = {};
    imageViewCreateInfo.sType                 = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    imageViewCreateInfo.image                 = depthImage;
    imageViewCreateInfo.viewType              = VK_IMAGE_VIEW_TYPE_2D;
    imageViewCreateInfo.format                = imageCreateInfo.format;
  VkComponentMapping g                      = { VK_COMPONENT_SWIZZLE_IDENTITY, VK_COMPONENT_SWIZZLE_IDENTITY, VK_COMPONENT_SWIZZLE_IDENTITY, VK_COMPONENT_SWIZZLE_IDENTITY };
    imageViewCreateInfo.components            = g; 
    imageViewCreateInfo.subresourceRange.aspectMask     = aspectMask;
    imageViewCreateInfo.subresourceRange.baseMipLevel   = 0;
    imageViewCreateInfo.subresourceRange.levelCount     = 1;
    imageViewCreateInfo.subresourceRange.baseArrayLayer = 0;
    imageViewCreateInfo.subresourceRange.layerCount    = 1;
    result = 
    vkCreateImageView( device, 
                   &imageViewCreateInfo, 
             NULL, 
             &depthImageView );
    DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create image view." );
  }
#endif // DEPTH_BUFFER

  // 0 - colour screen buffer
  VkAttachmentDescription pass[2] = { };
  pass[0].format                = VK_FORMAT_B8G8R8A8_UNORM;
  pass[0].samples               = VK_SAMPLE_COUNT_1_BIT;
  pass[0].loadOp                = VK_ATTACHMENT_LOAD_OP_CLEAR;
  pass[0].storeOp               = VK_ATTACHMENT_STORE_OP_STORE;
  pass[0].stencilLoadOp         = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
  pass[0].stencilStoreOp        = VK_ATTACHMENT_STORE_OP_DONT_CARE;
  pass[0].initialLayout         = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
  pass[0].finalLayout           = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

  VkAttachmentReference car     = {};
  car.attachment                = 0;
  car.layout                    = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

  // 1 - depth buffer
  pass[1].format              = VK_FORMAT_D16_UNORM;
  pass[1].samples             = VK_SAMPLE_COUNT_1_BIT;
  pass[1].loadOp              = VK_ATTACHMENT_LOAD_OP_CLEAR;
  pass[1].storeOp             = VK_ATTACHMENT_STORE_OP_DONT_CARE;
  pass[1].stencilLoadOp       = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
  pass[1].stencilStoreOp      = VK_ATTACHMENT_STORE_OP_DONT_CARE;
  pass[1].initialLayout       = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
  pass[1].finalLayout         = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
  
  // create the one main subpass of our renderpass:
  VkSubpassDescription subpass    = {};
  subpass.pipelineBindPoint       = VK_PIPELINE_BIND_POINT_GRAPHICS;
  subpass.colorAttachmentCount    = 1;
  subpass.pColorAttachments       = &car;
  subpass.pDepthStencilAttachment = NULL;
  
#ifdef DEPTH_BUFFER
  VkAttachmentReference dar    = {};
    dar.attachment          = 1;
    dar.layout            = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
  subpass.pDepthStencilAttachment = &dar;
#endif
  // create our main renderpass:
  VkRenderPassCreateInfo rpci    = {};
  rpci.sType            = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
  rpci.attachmentCount      = 1; // colour
#ifdef DEPTH_BUFFER
  rpci.attachmentCount      = 2; // colour and depth
#endif
  rpci.pAttachments        = pass;
  rpci.subpassCount        = 1;
  rpci.pSubpasses          = &subpass;

  VkResult result = 
  vkCreateRenderPass( device, 
            &rpci, 
            NULL, 
            outRenderPass );

  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create renderpass" );

#ifdef DEPTH_BUFFER
  VkImageView frameBufferAttachments[2] = {0};
#else
  VkImageView frameBufferAttachments[1] = {0};
#endif
  // create our frame buffers:
  VkFramebufferCreateInfo fbci  = {};
  fbci.sType            = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
  fbci.renderPass          = *outRenderPass;
  // must be equal to the attachment count on render pass
  fbci.attachmentCount      = 1;
#ifdef DEPTH_BUFFER
  fbci.attachmentCount      = 2;
#endif
  fbci.pAttachments        = frameBufferAttachments;
  fbci.width            = width;
  fbci.height            = height;
  fbci.layers            = 1;


  // create a framebuffer per swap chain imageView:
  *outFrameBuffers = new VkFramebuffer[ 2 ];
  for( uint32_t i = 0; i < 2; ++i ) 
  {
    frameBufferAttachments[0] = presentImageViews[ i ];
#ifdef DEPTH_BUFFER
    frameBufferAttachments[1] = depthImageView;
#endif
    result = 
    vkCreateFramebuffer( device, &fbci, NULL, &(*outFrameBuffers)[i] );
    DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to create framebuffer.");
  }// End for i
}// End SetupRenderPass(..)




/****************************************************************/
/*  
     Step 6
/*
/****************************************************************/

// Step -6-
void SetupCommandBuffer(VkDevice          device,
                        VkPhysicalDevice  physicalDevice,
                        VkCommandBuffer*  outCommandBuffer)
{
  // Give our device some commands (orders)
  {
  // we can now get the device queue we will be
  // submitting work to:
  VkQueue presentQueue = NULL;
  vkGetDeviceQueue( device, 0, 0, &presentQueue );

  // create our command buffers:
  VkCommandPoolCreateInfo cpci = {};
  cpci.sType            = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
  cpci.flags            = VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
  cpci.queueFamilyIndex = 0;

  VkCommandPool commandPool;
  VkResult result = 
  vkCreateCommandPool( device, 
                      &cpci, 
                      NULL, 
                      &commandPool );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create command pool." );

  VkCommandBufferAllocateInfo cbai  = {};
  cbai.sType                  = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
  cbai.commandPool            = commandPool;
  cbai.level                  = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
  cbai.commandBufferCount     = 1;

  result = 
  vkAllocateCommandBuffers( device, &cbai, outCommandBuffer );

  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to allocate draw command buffer." );
  }
}// End SetupCommandBuffer(..)









/****************************************************************/
/*  
     Step 7
/*
/****************************************************************/

// Step -7-
void SetupVertexBuffer(VkDevice           device,
                       VkPhysicalDevice   physicalDevice,
                       int*               outVertexSize,
                       int*               outNumberOfTriangles,
                       VkBuffer*          outVertexInputBuffer )
{
  // Our handle to our created vertex buffer and 
  // its data

    // Vertex info
  struct vertex 
  {
    float x, y, z, w; // position
    float r, g, b;    // color
    float u, v;       // texture
  };


  
  float r   = 20;
  float inc = 1.0f / (r);
  float s   = inc*0.5f;

  // Simple plane of triangles
  vector<vertex> tmpVertices;
  for (int x=0; x<r; ++x)
  {
    for (int z=0; z<r; ++z)
    {
      float dx = -0.5f + x*inc;
      float dz = -0.5f + z*inc;

      vertex quad[] = 
      {
        // x    y    z          w   r g b     u           v
        { -s+dx, 0.0f, -s+dz,   1,  1,0,0,   (x/r)    , (z/r)     },
        {  s+dx, 0.0f, -s+dz,   1,  1,0,0,   (x+1)/r  , (z/r)     },
        {  s+dx, 0.0f,  s+dz,   1,  1,0,0,   (x+1)/r  , (z+1)/r   },
        // triangle 2
        {  s+dx, 0.0f,  s+dz,   1,  0,1,0,   (x+1)/r  , (z+1)/r   }, 
        { -s+dx, 0.0f,  s+dz,   1,  0,1,0,   (x/r)    , (z+1)/r   }, 
        { -s+dx, 0.0f, -s+dz,   1,  0,1,0,   (x/r)    , (z/r)     }
      };

      for (int i=0; i<6; ++i) tmpVertices.push_back( quad[i] );
    }// End for z
  }// End for x


  *outVertexSize = sizeof(vertex);

  {
  const int numberOfVertices = tmpVertices.size();
  *outNumberOfTriangles = numberOfVertices / 3;

  DBG_ASSERT(*outNumberOfTriangles >0 && *outNumberOfTriangles<10000 );

  // create our vertex buffer:
  VkBufferCreateInfo vertexInputBufferInfo  = {};
  vertexInputBufferInfo.sType          = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
  // size in bytes of our data
  // a single triangle requires 3 vertices
  vertexInputBufferInfo.size                  = sizeof(vertex) * numberOfVertices; 
  vertexInputBufferInfo.usage                 = VK_BUFFER_USAGE_VERTEX_BUFFER_BIT;
  vertexInputBufferInfo.sharingMode           = VK_SHARING_MODE_EXCLUSIVE;
  vertexInputBufferInfo.queueFamilyIndexCount = 0;
  vertexInputBufferInfo.pQueueFamilyIndices   = NULL;

  VkResult result = 
  vkCreateBuffer( device, 
          &vertexInputBufferInfo, 
          NULL, 
          outVertexInputBuffer );  
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create vertex input buffer." );

  // allocate memory for buffers:
  VkMemoryRequirements vertexBufferMemoryRequirements = {};
  vkGetBufferMemoryRequirements( device, 
                   *outVertexInputBuffer, 
                   &vertexBufferMemoryRequirements );

  VkMemoryAllocateInfo bufferAllocateInfo = {};
  bufferAllocateInfo.sType        = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
  bufferAllocateInfo.pNext        = NULL;
  bufferAllocateInfo.allocationSize    = vertexBufferMemoryRequirements.size;

  // read the device memory properties
  VkPhysicalDeviceMemoryProperties memoryProperties;
    vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

  for( uint32_t i = 0; i <VK_MAX_MEMORY_TYPES; ++i ) 
  {
      VkMemoryType memoryType = memoryProperties.memoryTypes[i];
    // is this the memory type we are looking for?
      if( ( memoryType.propertyFlags & VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT ) ) 
      {
        // save location
        bufferAllocateInfo.memoryTypeIndex = i;
        break;
    }
  }// End for i

  VkDeviceMemory vertexBufferMemory;
  result = vkAllocateMemory( device, 
                            &bufferAllocateInfo, 
                            NULL, 
                            &vertexBufferMemory );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to allocate buffer memory." );

  // set buffer content:
  void *mapped = NULL;

  result = 
  vkMapMemory(device, 
        vertexBufferMemory, 
        0,
        VK_WHOLE_SIZE, 
        0, 
        &mapped );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to mapp buffer memory." );

  vertex *trianglevertices = (vertex *) mapped;

  // copy across your local data
  memcpy( trianglevertices, &tmpVertices[0], sizeof(vertex)*tmpVertices.size() );



  vkUnmapMemory(device, 
                vertexBufferMemory);

  result = 
  vkBindBufferMemory( device, 
                      *outVertexInputBuffer,
                      vertexBufferMemory, 
                      0 );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to bind buffer memory." );
  }
}// End SetupVertexBuffer(..)


/****************************************************************/
/*  
     Step 8
/*
/****************************************************************/

// Step -8-
void SetupShaderandUniforms(VkDevice          device,
                            VkPhysicalDevice  physicalDevice,
                            int               width,
                            int               height,
                            VkShaderModule*   outVertShaderModule,
                            VkShaderModule*   outFragShaderModule,
                            VkShaderModule*   outGeomShaderModule,
                            VkBuffer*         outBuffer,
                            VkDeviceMemory*   outMemory)
{

  {
  // Simple Shaders (Vertex & Fragement)
  uint32_t codeSize = 0;
  // assume your file is less than 10,000 bytes
  char *code        = new char[10000];
  FILE* fileHandle = NULL;

  
  // #############
  // Vertex Shader
  // #############

  // load your vertex shader:
  fileHandle = fopen( ".\\vert.spv", "rb" );

  // did we successfully find file
  DBG_ASSERT_MSG(fileHandle != NULL,
    "Failed to open shader file.");
  
  // read the file contents
  codeSize = fread(code, 1, 10000, fileHandle);
  // close the file
  fclose(fileHandle);
  fileHandle = NULL;

  VkShaderModuleCreateInfo vertexShaderCreationInfo = {};
  vertexShaderCreationInfo.sType     = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
  vertexShaderCreationInfo.codeSize  = codeSize;
  vertexShaderCreationInfo.pCode     = (uint32_t *)code;

  VkResult result = 
  vkCreateShaderModule( device, 
                        &vertexShaderCreationInfo, 
                        NULL, 
                        outVertShaderModule );
  DBG_ASSERT_VULKAN_MSG( result, "Failed to create vertex shader module." );


  // #############
  // Fragment Shader
  // #############

  // load your fragment shader:
  fileHandle = fopen( ".\\frag.spv", "rb" );

  // did we successfully find file
  DBG_ASSERT_MSG(fileHandle != NULL,
    "Failed to open shader file.");
  
  // read the file contents
  codeSize = fread(code, 1, 10000, fileHandle);
  // close the file
  fileHandle = NULL;

  VkShaderModuleCreateInfo fragmentShaderCreationInfo = {};
  fragmentShaderCreationInfo.sType    = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
  fragmentShaderCreationInfo.codeSize = codeSize;
  fragmentShaderCreationInfo.pCode    = (uint32_t *)code;

  result = 
  vkCreateShaderModule( device, 
                        &fragmentShaderCreationInfo, 
                        NULL, 
                        outFragShaderModule );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create fragment shader module." );


  // #############
  // Geometry Shader
  // #############
  // load our geometry shader:
  fileHandle = fopen( ".\\geom.spv", "rb" );

  // did we successfully find file
  DBG_ASSERT_MSG(fileHandle != NULL,
    "Failed to open shader file.");
  
  // read the file contents
  codeSize = fread(code, 1, 10000, fileHandle);
  // close the file
  fileHandle = NULL;

  VkShaderModuleCreateInfo geometryShaderCreationInfo = {};
  geometryShaderCreationInfo.sType    = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
  geometryShaderCreationInfo.codeSize = codeSize;
  geometryShaderCreationInfo.pCode    = (uint32_t *)code;

  result = 
  vkCreateShaderModule( device, 
                        &geometryShaderCreationInfo, 
                        NULL, 
                        outGeomShaderModule );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create geometry shader module." );


  delete code;
  code = NULL;
  }




  {
  // Create `uniform' buffer for passing constant
  // data to the shader (connecting shader with the data)
  const float PI      = 3.14159265359f;
  const float TORAD   = PI/180.0f;

  // perspective projection parameters:
  float fov           = 45.0f;
  float nearZ         = 0.1f;
  float farZ          = 1000.0f;
  float aspectRatio   = width / (float)height;
  float t             = 1.0f / tanf( fov * TORAD * 0.5f );
  float nf            = nearZ - farZ;

  // simple matrices (model/view/projection)
  static
  float lprojectionMatrix[16] = 
  { t / aspectRatio, 0,    0,                  0,
    0,               t,    0,                  0,
    0,               0,   (-nearZ-farZ) / nf, (2*nearZ*farZ) / nf,
    0,               0,    1,                  0 };

  static
  float lviewMatrix[16] = 
  {  1, 0, 0, 0,
     0, 1, 0, 0,
     0, 0, 1, 0,
     0, 0, 0, 1 };

  float xp =   0;
  float yp =   0.3f;
  float zp =   1.5;

  static
  float lmodelMatrix[16] = 
  {   1, 0, 0, xp,
      0, 1, 0, yp,
      0, 0, 1, zp,
      0, 0, 0, 1 };

  // create our uniforms buffers:
  VkBufferCreateInfo bufferCreateInfo = {};
  bufferCreateInfo.sType          = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
  // size in bytes
  bufferCreateInfo.size           = sizeof(float) * 16 * 3; 
  bufferCreateInfo.usage          = VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT;
  bufferCreateInfo.sharingMode    = VK_SHARING_MODE_EXCLUSIVE;

  VkResult result = 
  vkCreateBuffer( device, 
                  &bufferCreateInfo, 
                  NULL,  
                  outBuffer );  
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create uniforms buffer." );

  // allocate memory for buffer:
  VkMemoryRequirements bufferMemoryRequirements = {};
  vkGetBufferMemoryRequirements( device, *outBuffer, &bufferMemoryRequirements );

  VkMemoryAllocateInfo matrixAllocateInfo = {};
  matrixAllocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
  matrixAllocateInfo.allocationSize = bufferMemoryRequirements.size;

  VkPhysicalDeviceMemoryProperties memoryProperties;
    vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

  for( uint32_t i = 0; i <VK_MAX_MEMORY_TYPES; ++i ) 
  {
      VkMemoryType memoryType = memoryProperties.memoryTypes[i];
    // is this the memory type we are looking for?
      if( ( memoryType.propertyFlags & VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT ) ) 
      {
        // save location
        matrixAllocateInfo.memoryTypeIndex = i;
        break;
    }
  }

  result = 
  vkAllocateMemory( device, &matrixAllocateInfo, NULL, outMemory );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to allocate uniforms buffer memory." );

  result = 
  vkBindBufferMemory( device, *outBuffer, *outMemory, 0 );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to bind uniforms buffer memory." );

  // set buffer content:
  void *matrixMapped = NULL;
  result = vkMapMemory( device, *outMemory, 0, VK_WHOLE_SIZE, 0, &matrixMapped );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to mapp uniform buffer memory." );

  memcpy( matrixMapped, &lprojectionMatrix[0], sizeof( lprojectionMatrix ) );
  memcpy( ((float *)matrixMapped + 16), &lviewMatrix[0], sizeof( lviewMatrix ) );
  memcpy( ((float *)matrixMapped + 32), &lmodelMatrix[0], sizeof( lmodelMatrix ) );

  vkUnmapMemory( device, *outMemory );
  }
}// End SetupShaderandUniforms(..)


/****************************************************************/
/*  
     Step 9
/*
/****************************************************************/

// Step -9-
void SetupDescriptors(VkDevice                device,
                      VkPhysicalDevice        physicalDevice,
                      VkBuffer                buffer,
                      VkDescriptorSet*        outDescriptorSet,
                      VkDescriptorSetLayout*  outDescriptorSetLayout)
{
  {
  // Define and create our descriptors:
  VkDescriptorSetLayoutBinding bindings[2];

  // uniform buffer for our matrices:
  bindings[0].binding               = 0;
  bindings[0].descriptorType        = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
  bindings[0].descriptorCount       = 1;
  bindings[0].stageFlags            = VK_SHADER_STAGE_VERTEX_BIT |
                                      VK_SHADER_STAGE_GEOMETRY_BIT;
  bindings[0].pImmutableSamplers    = NULL;

  // uniform buffer for our matrices:
  bindings[1].binding               = 1;
  bindings[1].descriptorType        = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
  bindings[1].descriptorCount       = 1;
  bindings[1].stageFlags            = VK_SHADER_STAGE_FRAGMENT_BIT |
                                      VK_SHADER_STAGE_VERTEX_BIT;
  bindings[1].pImmutableSamplers    = NULL;

  VkDescriptorSetLayoutCreateInfo setLayoutCreateInfo = {};
  setLayoutCreateInfo.sType         = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
  setLayoutCreateInfo.bindingCount  = 2;
  setLayoutCreateInfo.pBindings     = bindings;


  VkResult result = 
  vkCreateDescriptorSetLayout( device, 
                               &setLayoutCreateInfo, 
                               NULL, 
                               outDescriptorSetLayout );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create DescriptorSetLayout." );

  // decriptor pool creation:
  VkDescriptorPoolSize uniformBufferPoolSize[2];
  uniformBufferPoolSize[0].type             = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
  uniformBufferPoolSize[0].descriptorCount  = 1;

  uniformBufferPoolSize[1].type             = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
  uniformBufferPoolSize[1].descriptorCount  = 1;

  VkDescriptorPoolCreateInfo poolCreateInfo  = {}; 
  poolCreateInfo.sType                = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
  poolCreateInfo.maxSets              = 1;
  poolCreateInfo.poolSizeCount        = 1;
  poolCreateInfo.pPoolSizes           = uniformBufferPoolSize;


  VkDescriptorPool descriptorPool;
  result = vkCreateDescriptorPool( device, 
                                   &poolCreateInfo, 
                                   NULL, 
                                   &descriptorPool );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create descriptor pool." );

  // allocate our descriptor from the pool:
  VkDescriptorSetAllocateInfo dsai  = {};
  dsai.sType                        = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
  dsai.descriptorPool               = descriptorPool;
  dsai.descriptorSetCount           = 1;
  dsai.pSetLayouts                  = outDescriptorSetLayout;

  result = 
  vkAllocateDescriptorSets( device, 
                            &dsai, 
                            outDescriptorSet );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to allocate descriptor sets." );

  // When a set is allocated all values are undefined
  // and all descriptors are uninitialized. We must 
  // init all statically used bindings:
  VkDescriptorBufferInfo dbi  = {};
  dbi.buffer                  = buffer;
  dbi.offset                  = 0;
  dbi.range                   = VK_WHOLE_SIZE;


  VkWriteDescriptorSet wd[2]   = {};
  wd[0].sType                  = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
  wd[0].dstSet                 = *outDescriptorSet;
  wd[0].dstBinding             = 0;
  wd[0].dstArrayElement        = 0;
  wd[0].descriptorCount        = 1;
  wd[0].descriptorType         = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
  wd[0].pImageInfo             = NULL;
  wd[0].pBufferInfo            = &dbi;
  wd[0].pTexelBufferView       = NULL;


  // Texture
  Texture* texture = new Texture(device, physicalDevice,"pattern.tga");

  wd[1].sType                  = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
  wd[1].dstSet                 = *outDescriptorSet;
  wd[1].dstBinding             = 1;
  wd[1].dstArrayElement        = 0;
  wd[1].descriptorCount        = 1;
  wd[1].descriptorType         = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
  wd[1].pImageInfo             = &texture->descriptorImageInfo;
  wd[1].pBufferInfo            = NULL;
  wd[1].pTexelBufferView       = NULL;

  vkUpdateDescriptorSets( device, 
                          2, 
                          wd, 
                          0, 
                          NULL );
  }
}// End SetupDescriptors(..)



/****************************************************************/
/*  
     Step 10
/*
/****************************************************************/

// Step -10-
void SetupPipeline( VkDevice                device,
                    int                     width,
                    int                     height,
                    int                     vertexSize,
                    VkDescriptorSetLayout   descriptorSetLayout,
                    VkShaderModule          vertShaderModule,
                    VkShaderModule          fragShaderModule,
                    VkShaderModule          geomShaderModule,
                    VkRenderPass            renderPass, 
                    VkPipeline*             outPipeline,
                    VkPipelineLayout*       outPipelineLayout)
{
  {
  // Graphics Pipeline:
  VkPipelineLayoutCreateInfo layoutCreateInfo   = {};
  layoutCreateInfo.sType                        = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
  layoutCreateInfo.setLayoutCount               = 1;
  layoutCreateInfo.pSetLayouts                  = &descriptorSetLayout;
  layoutCreateInfo.pushConstantRangeCount       = 0;
  layoutCreateInfo.pPushConstantRanges          = NULL;

  VkResult result = 
  vkCreatePipelineLayout( device, 
              &layoutCreateInfo, 
              NULL, 
              outPipelineLayout );
  DBG_ASSERT_VULKAN_MSG( result, "Failed to create pipeline layout." );



  // setup shader stages in the pipeline:
  VkPipelineShaderStageCreateInfo shaderStageCreateInfo[3] = {};
  shaderStageCreateInfo[0].sType                = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
  shaderStageCreateInfo[0].stage                = VK_SHADER_STAGE_VERTEX_BIT;
  shaderStageCreateInfo[0].module               = vertShaderModule;
  // shader entry point function name
  shaderStageCreateInfo[0].pName                = "main";
  shaderStageCreateInfo[0].pSpecializationInfo  = NULL;

  shaderStageCreateInfo[1].sType                = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
  shaderStageCreateInfo[1].stage                = VK_SHADER_STAGE_FRAGMENT_BIT;
  shaderStageCreateInfo[1].module               = fragShaderModule;
  // shader entry point function name
  shaderStageCreateInfo[1].pName                = "main";
  shaderStageCreateInfo[1].pSpecializationInfo  = NULL;

  shaderStageCreateInfo[2].sType                = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
  shaderStageCreateInfo[2].stage                = VK_SHADER_STAGE_GEOMETRY_BIT;
  shaderStageCreateInfo[2].module               = geomShaderModule;
  // shader entry point function name
  shaderStageCreateInfo[2].pName                = "main";
  shaderStageCreateInfo[2].pSpecializationInfo  = NULL;

  // vertex input configuration:
  VkVertexInputBindingDescription vertexBindingDescription = {};
  vertexBindingDescription.binding            = 0;
  vertexBindingDescription.stride             = vertexSize;
  vertexBindingDescription.inputRate          = VK_VERTEX_INPUT_RATE_VERTEX;

  VkVertexInputAttributeDescription vertexAttributeDescritpion[3];

  // position:
  vertexAttributeDescritpion[0].location      = 0;
  vertexAttributeDescritpion[0].binding       = 0;
  vertexAttributeDescritpion[0].format        = VK_FORMAT_R32G32B32A32_SFLOAT;
  vertexAttributeDescritpion[0].offset        = 0;

  // colors:
  vertexAttributeDescritpion[1].location      = 1;
  vertexAttributeDescritpion[1].binding       = 0;
  vertexAttributeDescritpion[1].format        = VK_FORMAT_R32G32B32_SFLOAT;
  vertexAttributeDescritpion[1].offset        = 4 * sizeof(float);


  vertexAttributeDescritpion[2].location      = 2;
  vertexAttributeDescritpion[2].binding       = 0;
  vertexAttributeDescritpion[2].format        = VK_FORMAT_R32G32_SFLOAT;
  vertexAttributeDescritpion[2].offset        = 7 * sizeof(float);

  VkPipelineVertexInputStateCreateInfo vertexInputStateCreateInfo = {};
  vertexInputStateCreateInfo.sType                            = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
  vertexInputStateCreateInfo.vertexBindingDescriptionCount    = 1;
  // attribute indexing is a function of the vertex index
  vertexInputStateCreateInfo.pVertexBindingDescriptions       = &vertexBindingDescription;  
  vertexInputStateCreateInfo.vertexAttributeDescriptionCount  = 3;
  vertexInputStateCreateInfo.pVertexAttributeDescriptions     = vertexAttributeDescritpion;

  // vertex topology config:
  VkPipelineInputAssemblyStateCreateInfo inputAssemblyStateCreateInfo = {};
  inputAssemblyStateCreateInfo.sType              = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
  inputAssemblyStateCreateInfo.topology           = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
  //inputAssemblyStateCreateInfo.topology           = VK_PRIMITIVE_TOPOLOGY_LINE_LIST;
  inputAssemblyStateCreateInfo.primitiveRestartEnable = VK_FALSE;

  // viewport config:
  VkViewport viewport               = {};
  viewport.x                        = 0;
  viewport.y                        = 0;
  viewport.width                    = (float)width;
  viewport.height                   = (float)height;
  viewport.minDepth                 = 0;
  viewport.maxDepth                 = 1;

  VkRect2D scissors                 = {};
  VkOffset2D k                      = {0,0};
  VkExtent2D m                      = {width,height };
  scissors.offset                   = k;
  scissors.extent                   = m;

  VkPipelineViewportStateCreateInfo viewportState = {};
  viewportState.sType                     = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
  viewportState.viewportCount             = 1;
  viewportState.pViewports                = &viewport;
  viewportState.scissorCount              = 1;
  viewportState.pScissors                 = &scissors;

  // rasterization config:
  VkPipelineRasterizationStateCreateInfo rasterizationState = {};
  rasterizationState.sType                  = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
  rasterizationState.depthClampEnable       = VK_FALSE;
  rasterizationState.rasterizerDiscardEnable= VK_FALSE;
  rasterizationState.polygonMode            = VK_POLYGON_MODE_FILL;
  rasterizationState.cullMode               = VK_CULL_MODE_NONE;
  rasterizationState.frontFace              = VK_FRONT_FACE_COUNTER_CLOCKWISE;
  rasterizationState.depthBiasEnable        = VK_FALSE;
  rasterizationState.depthBiasConstantFactor= 0;
  rasterizationState.depthBiasClamp         = 0;
  rasterizationState.depthBiasSlopeFactor   = 0;
  rasterizationState.lineWidth              = 1;

  // sampling config:
  VkPipelineMultisampleStateCreateInfo multisampleState = {};
  multisampleState.sType =
  VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
  multisampleState.rasterizationSamples     = VK_SAMPLE_COUNT_1_BIT;
  multisampleState.sampleShadingEnable      = VK_FALSE;
  multisampleState.minSampleShading         = 0;
  multisampleState.pSampleMask              = NULL;
  multisampleState.alphaToCoverageEnable    = VK_FALSE;
  multisampleState.alphaToOneEnable         = VK_FALSE;



  // color blend config: (Actually off for tutorial)
  VkPipelineColorBlendAttachmentState colorBlendAttachmentState = {};
  colorBlendAttachmentState.blendEnable           = VK_FALSE;
  colorBlendAttachmentState.srcColorBlendFactor   = VK_BLEND_FACTOR_SRC_COLOR;
  colorBlendAttachmentState.dstColorBlendFactor   = VK_BLEND_FACTOR_ONE_MINUS_DST_COLOR;
  colorBlendAttachmentState.colorBlendOp          = VK_BLEND_OP_ADD;
  colorBlendAttachmentState.srcAlphaBlendFactor   = VK_BLEND_FACTOR_ZERO;
  colorBlendAttachmentState.dstAlphaBlendFactor   = VK_BLEND_FACTOR_ZERO;
  colorBlendAttachmentState.alphaBlendOp          = VK_BLEND_OP_ADD;
  colorBlendAttachmentState.colorWriteMask        = 0xf;

  VkPipelineColorBlendStateCreateInfo colorBlendState = {};
  colorBlendState.sType                       = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
  colorBlendState.logicOpEnable               = VK_FALSE;
  colorBlendState.logicOp                     = VK_LOGIC_OP_CLEAR;
  colorBlendState.attachmentCount             = 1;
  colorBlendState.pAttachments                = &colorBlendAttachmentState;
  colorBlendState.blendConstants[0]           = 0.0;
  colorBlendState.blendConstants[1]           = 0.0;
  colorBlendState.blendConstants[2]           = 0.0;
  colorBlendState.blendConstants[3]           = 0.0;

  // configure dynamic state:
  VkDynamicState dynamicState[2]              = { VK_DYNAMIC_STATE_VIEWPORT, VK_DYNAMIC_STATE_SCISSOR };
  VkPipelineDynamicStateCreateInfo dynamicStateCreateInfo = {};
  dynamicStateCreateInfo.sType                = VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
  dynamicStateCreateInfo.dynamicStateCount    = 2;
  dynamicStateCreateInfo.pDynamicStates       = dynamicState;

  // add depth buffer options to the pipeline
  VkPipelineDepthStencilStateCreateInfo depthStencil = {};
  depthStencil.sType                          = VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
  depthStencil.depthTestEnable                = VK_TRUE;
  depthStencil.depthWriteEnable               = VK_TRUE;
  depthStencil.depthCompareOp                 = VK_COMPARE_OP_LESS;
  depthStencil.depthBoundsTestEnable          = VK_FALSE;
  depthStencil.minDepthBounds                 = 0.0f; // Optional
  depthStencil.maxDepthBounds                 = 1.0f; // Optional
  depthStencil.stencilTestEnable              = VK_FALSE;


  // and finally, pipeline config and creation:
  VkGraphicsPipelineCreateInfo pipelineCreateInfo = {};
  pipelineCreateInfo.sType                    = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
  pipelineCreateInfo.stageCount               = 3;
  pipelineCreateInfo.pStages                  = shaderStageCreateInfo;
  pipelineCreateInfo.pVertexInputState        = &vertexInputStateCreateInfo;
  pipelineCreateInfo.pInputAssemblyState      = &inputAssemblyStateCreateInfo;
  pipelineCreateInfo.pTessellationState       = NULL;
  pipelineCreateInfo.pViewportState           = &viewportState;
  pipelineCreateInfo.pRasterizationState      = &rasterizationState;
  pipelineCreateInfo.pMultisampleState        = &multisampleState;
  pipelineCreateInfo.pDepthStencilState       = &depthStencil;
  pipelineCreateInfo.pColorBlendState         = &colorBlendState;
  pipelineCreateInfo.pDynamicState            = &dynamicStateCreateInfo;
  pipelineCreateInfo.layout                   = *outPipelineLayout;
  pipelineCreateInfo.renderPass               = renderPass;
  pipelineCreateInfo.subpass                  = 0;
  pipelineCreateInfo.basePipelineHandle       = NULL;
  pipelineCreateInfo.basePipelineIndex        = 0;

  result = 
  vkCreateGraphicsPipelines( device, 
                 VK_NULL_HANDLE, 
                 1, 
                 &pipelineCreateInfo, 
                 NULL, 
                 outPipeline );

  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create graphics pipeline." );
  }
}// End SetupPipeline(..)



/****************************************************************/
/*  
     Step 11
/*
/****************************************************************/

// Step -11-
void RenderLoop(VkDevice        device,
                int             width,
                int             height,
                int             numberOfTriangles,
                VkSwapchainKHR  swapChain,
                VkCommandBuffer commandBuffer,
                VkImage*        presentImages,
                VkFramebuffer*  frameBuffers,
                VkRenderPass    renderPass,
                VkBuffer        vertexInputBuffer,
                VkDescriptorSet descriptorSet,
                VkPipelineLayout pipelineLayout,
                VkPipeline      pipeline)
{
  
#if 1 
  uint32_t nextImageIdx;
  vkAcquireNextImageKHR(  device, 
              swapChain, 
              UINT64_MAX,
              VK_NULL_HANDLE, 
              VK_NULL_HANDLE, 
              &nextImageIdx );

  VkCommandBufferBeginInfo beginInfo  = {};
  beginInfo.sType            = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
  beginInfo.flags            = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;

  vkBeginCommandBuffer(commandBuffer, 
             &beginInfo );

  // activate render pass:
  VkClearValue clearValue[]      = { { 1.0f, 1.0f, 1.0f, 1.0f }, { 1.0, 0.0 } };
  VkRenderPassBeginInfo renderPassBeginInfo = {};
  renderPassBeginInfo.sType      = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
  renderPassBeginInfo.renderPass    = renderPass;
  renderPassBeginInfo.framebuffer    = frameBuffers[ nextImageIdx ];
  VkOffset2D a            = {0, 0}; 
  VkExtent2D b            = {width, height};
  VkRect2D   c            = {a,b}; 
  renderPassBeginInfo.renderArea      = c;
  renderPassBeginInfo.clearValueCount = 2;
  renderPassBeginInfo.pClearValues    = clearValue;
  vkCmdBeginRenderPass(  commandBuffer, 
              &renderPassBeginInfo, 
              VK_SUBPASS_CONTENTS_INLINE );

  // bind the graphics pipeline to the command buffer. 
  // Any vkDraw command afterwards is affected by this
  // pipeline.
  vkCmdBindPipeline( commandBuffer, 
            VK_PIPELINE_BIND_POINT_GRAPHICS, 
            pipeline );    

  // take care of dynamic state:
  VkViewport viewport = { 0, 0, (float)width, (float)height, 0, 1 };
  vkCmdSetViewport( commandBuffer, 0, 1, &viewport );

  VkRect2D scissor = { 0, 0, width, height };
  vkCmdSetScissor( commandBuffer, 0, 1, &scissor);

  // bind our shader parameters:
  vkCmdBindDescriptorSets( commandBuffer, 
               VK_PIPELINE_BIND_POINT_GRAPHICS, 
               pipelineLayout, 
               0, 
               1, 
               &descriptorSet, 
               0, 
               NULL );

  // render the triangle:
  VkDeviceSize offsets = { 0 };
  vkCmdBindVertexBuffers( commandBuffer, 
              0, 
              1, 
              &vertexInputBuffer, 
              &offsets );

  vkCmdDraw( commandBuffer, 
         // vertex count
         numberOfTriangles * 3, 
         // instance count
         1, 
         // first index
         0, 
         // first instance
         0 );

  vkCmdEndRenderPass( commandBuffer );

  vkEndCommandBuffer( commandBuffer );


  // present:
  VkFence renderFence;
  VkFenceCreateInfo fenceCreateInfo  = {};
  fenceCreateInfo.sType        = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
  vkCreateFence( device, 
           &fenceCreateInfo, 
           NULL, 
           &renderFence );

  VkPipelineStageFlags waitStageMash = { VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT };

  VkQueue presentQueue = 0;
  vkGetDeviceQueue( device, 
            0, 
            0, 
            &presentQueue );

  VkSubmitInfo submitInfo      = {};
  submitInfo.sType        = VK_STRUCTURE_TYPE_SUBMIT_INFO;
  submitInfo.waitSemaphoreCount  = 0;
  submitInfo.pWaitSemaphores    = VK_NULL_HANDLE;
  submitInfo.pWaitDstStageMask    = &waitStageMash;
  submitInfo.commandBufferCount   = 1;
  submitInfo.pCommandBuffers      = &commandBuffer;
  submitInfo.signalSemaphoreCount = 0;
  submitInfo.pSignalSemaphores    = VK_NULL_HANDLE;

  // Submit command buffer (drawing)
  vkQueueSubmit(  presentQueue, 
          1, 
          &submitInfo, 
          renderFence );

  // Wait for the the command/queue to finish
  vkWaitForFences( device, 1, &renderFence, VK_TRUE, UINT64_MAX );
  vkDestroyFence( device, renderFence, NULL );

  // Present the backbuffer
  VkPresentInfoKHR presentInfo  = {};
  presentInfo.sType        = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR;
  presentInfo.pNext        = NULL;
  presentInfo.waitSemaphoreCount  = 0;
  presentInfo.pWaitSemaphores    = VK_NULL_HANDLE;
  presentInfo.swapchainCount    = 1;
  presentInfo.pSwapchains      = &swapChain;
  presentInfo.pImageIndices    = &nextImageIdx;
  presentInfo.pResults      = NULL;

  vkQueuePresentKHR( presentQueue, 
            &presentInfo );

#endif

}// End RenderLoop(..)











